importScripts("/lib/ComPDFkit.js")

import Color from '../color';
import { parseAdobePDFTimestamp, convertToPDFTimestamp, roundToDecimalPlaces } from '../ui_utils';
import { AnnotationType, WidgetType, LineTypeString, StampType, StampTypeString, TextStampColorString, TextStampShapeString, ActionTypeString, WidgetTypeString, BorderStyleString, CheckStyleString, ALIGN } from '../../constants'

let ComPDFKitJS = {}
let DataArray = []
let LineType = {}
let BorderStyle = {}
let PageSize = {}
let Rect = {}
let RGBColor = {}
let FontDa = {}
let PDFDestination = {}
let WidgetItem = {}
let WidgetItemsArray = {}
let EditCharPlace = {}
let EndEditCharPlace = {}
let RectArray = []
let CursorPoints = {}
let EditTextStyle = {}

import MessageHandler from "../message_handler"

async function sleep(delay) {
  await new Promise((resolve) => setTimeout(resolve, delay))
}

class CPDFWorker {
  static async setup(doc) {
    this.doc = doc
    const messageHandler = new MessageHandler('worker', 'main', self)
    this.MessageHandler = messageHandler

    let testMessageProcessed = false
    
    messageHandler.on("test", function (data) {
      if (testMessageProcessed) {
        return; // we already processed 'test' message once
      }
      testMessageProcessed = true;

      // Ensure that `TypedArray`s can be sent to the worker.
      messageHandler.send("test", data instanceof Uint8Array);
    });

    messageHandler.send("ready", null)

    messageHandler.send("initDoc", { doc })

    messageHandler.on("Ready", function (data) {
      return CPDFWorker.createDocumentHandler(self)
    })
  }

  static createDocumentHandler(port) {
    const messageHandler = this.MessageHandler
    
    messageHandler.on('LoadFile', (data) => {
      ComPDFKitJS.opened_files = []
      ComPDFKitJS.opened_files[0] = data.buffer
      return true
    })

    messageHandler.on('LoadDocumentByStream', (data) => {
      const { doc, fileId, length, password: rawPassword } = data
      const password = stringToNewUTF8(rawPassword)
      return Module._LoadDocumentByStream(doc, fileId, length, password)
    })

    messageHandler.on('GetPageCount', (data) => {
      const { doc } = data
      return Module._GetPageCount(doc)
    })

    messageHandler.on('GetPageSize', (data) => {
      const { doc, pageIndex } = data

      PageSize.Width = 0
      PageSize.Height = 0

      Module._GetPageSizeForIndex(doc, pageIndex)
      const pagePtr = Module._GetPage(doc, pageIndex)

      return {
        pagePtr,
        width: PageSize.Width,
        height: PageSize.Height
      }
    })

    messageHandler.on('InitAnnot', (data) => {
      const { pagePtr } = data
      const annotCount = Module._GetAnnotCount(pagePtr)
      const annotations = []
      for (let i = 0; i < annotCount; i++) {
        const annotPtr = Module._InitAnnot(pagePtr, i)
        const type = Module._GetAnnotType(annotPtr)
        const annotation = {
          annotPtr,
          type,
          pagePtr
        }
        annotations.push(annotation)
      }

      return {
        annotCount,
        annotations
      }
    })

    messageHandler.on('GetAnnotation', (data) => {
      const { doc, pagePtr, annotPtr, typeInt, scale } = data

      return getAnnotation({
        doc,
        pagePtr,
        annotPtr,
        typeInt,
        scale
      })
    })

    messageHandler.on('GetWidgetAnnotation', (data) => {
      const { doc, pagePtr, annotPtr, typeInt } = data

      return getWidgetAnnotation({
        doc,
        pagePtr,
        annotPtr,
        typeInt
      })
    })

    messageHandler.on('GetAnnotText', (data) => {
      const { pagePtr, annotPtr } = data

      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))
      console.log('content', content, createDate)

      FontDa = {}
      Module._GetAnnotFont(annotPtr)

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        font: FontDa.fontSize
      }
    })

    messageHandler.on('GetAnnotLink', (data) => {
      const { pagePtr, annotPtr } = data

      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const date = UTF8ToString(rawCreateDate)
      const createDate = date && parseAdobePDFTimestamp(date)

      FontDa = {}
      Module._GetAnnotFont(annotPtr)

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        font: FontDa.fontSize
      }
    })

    messageHandler.on('GetAnnotFreeText', (data) => {
      const { pagePtr, annotPtr } = data
      
      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }
      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      FontDa = {}
      Module._GetAnnotFont(annotPtr)

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        font: FontDa.fontSize
      }
    })

    messageHandler.on('GetAnnotLine', (data) => {
      const { pagePtr, annotPtr } = data

      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const date = UTF8ToString(rawCreateDate)
      const createDate = date && parseAdobePDFTimestamp(date)

      const lineWidth = Module._GetAnnotBorderWidth(annotPtr)

      RGBColor = {}
      const borderColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        borderColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      const transparency = Module._GetAnnotTransparency(annotPtr)

      LineType = {}
      Module._GetLineType(annotPtr)

      LinePoint = {}
      Module._GetLinePoint(pagePtr, annotPtr)
      const linePoints = [LinePoint.StartPointX, LinePoint.StartPointY, LinePoint.EndPointX, LinePoint.EndPointY]

      return {
        createDate,
        content,
        rect,
        borderColor,
        lineWidth,
        linePoints,
        transparency,
        tail: LineTypeString[LineType.Tail],
        head: LineTypeString[LineType.Head]
      }
    })

    messageHandler.on('GetAnnotSquare', (data) => {
      const { pagePtr, annotPtr, typeInt } = data

      getAnnotation({
        pagePtr,
        annotPtr,
        typeInt
      })

      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))
      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)

      const lineWidth = Module._GetAnnotBorderWidth(annotPtr)

      const transparency = Module._GetAnnotTransparency(annotPtr)
      const fillTransparency = Module._GetAnnotFilledTransparency(annotPtr)
      
      const annotation = {
        createDate,
        content,
        rect,
        lineWidth,
        transparency,
        fillTransparency
      }

      RGBColor = {}
      const isFilled = Module._GetAnnotFilledRGBColor(annotPtr)
      if (isFilled) {
        const fillColor = {}
        for (key in RGBColor) {
          fillColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }
        annotation.fillColor = fillColor
      }

      RGBColor = {}
      const hasColor = Module._GetAnnotRGBColor(annotPtr)
      if (hasColor) {
        const borderColor = {}
        for (key in RGBColor) {
          borderColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }

        annotation.borderColor = borderColor
      }

      return annotation
    })

    messageHandler.on('GetAnnotCircle', (data) => {
      const { pagePtr, annotPtr } = data
      
      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      FontDa = {}
      Module._GetAnnotFont(annotPtr)

      const lineWidth = Module._GetAnnotBorderWidth(annotPtr)

      RGBColor = {}
      const fillColor = {}
      Module._GetAnnotFilledRGBColor(annotPtr)
      for (key in RGBColor) {
        fillColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      RGBColor = {}
      const borderColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        borderColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      const transparency = Module._GetAnnotTransparency(annotPtr)

      return {
        content,
        createDate,
        rect,
        fillColor,
        lineWidth,
        borderColor,
        transparency
      }
    })
    
    messageHandler.on('GetAnnotPolygon', (data) => {
      const { pagePtr, annotPtr } = data

      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      FontDa = {}
      Module._GetAnnotFont(annotPtr)

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        font: FontDa.fontSize
      }
    })
    
    messageHandler.on('GetAnnotPolyline', (data) => {
      const { pagePtr, annotPtr } = data
      
      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      FontDa = {}
      Module._GetAnnotFont(annotPtr)

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        font: FontDa.fontSize
      }
    })
    
    messageHandler.on('GetAnnotHighlight', (data) => {
      const { pagePtr, annotPtr } = data
      
      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      InkPointArray = []
      Module._GetQuadPoints(pagePtr, annotPtr)

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        quadPoints: InkPointArray
      }
    })
    
    messageHandler.on('GetAnnotUnderline', (data) => {
      const { pagePtr, annotPtr } = data
      
      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      InkPointArray = []
      Module._GetQuadPoints(pagePtr, annotPtr)

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        quadPoints: InkPointArray
      }
    })
    
    messageHandler.on('GetAnnotSquiggly', (data) => {
      const { pagePtr, annotPtr } = data
      
      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      InkPointArray = []
      Module._GetQuadPoints(pagePtr, annotPtr)

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        quadPoints: InkPointArray
      }
    })
    
    messageHandler.on('GetAnnotStrikeout', (data) => {
      const { pagePtr, annotPtr } = data
      
      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      InkPointArray = []
      Module._GetQuadPoints(pagePtr, annotPtr)

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        quadPoints: InkPointArray
      }
    })
    
    messageHandler.on('GetAnnotStamp', (data) => {
      const { pagePtr, annotPtr } = data

      Rect = {}
      Module._GetAnnotRect(pagePtr, annotPtr)
      const rect = {
        left: roundToDecimalPlaces(Rect.Left),
        top: roundToDecimalPlaces(Rect.Top),
        right: roundToDecimalPlaces(Rect.Right),
        bottom: roundToDecimalPlaces(Rect.Bottom)
      }

      const rawContent = Module._GetAnnotContent(annotPtr)
      const content = UTF8ToString(rawContent)
      const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
      const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))

      FontDa = {}
      Module._GetAnnotFont(annotPtr)

      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      for (key in RGBColor) {
        RGBColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
      }

      return {
        content,
        createDate,
        rect,
        color: RGBColor,
        font: FontDa.fontSize
      }
    })

    messageHandler.on('PushRenderTask', (data) => {
      const { pagePtr, scale, left, right, bottom, top } = data
      
      let param = {
        pagePtr,
        scale,
        left,
        right,
        bottom,
        top,
        mode: -1,
        flag: 0,
        form: false
      }
      let imageArray = RenderPageBitmapWithMatrix(param)
      return { imageArray }
    })

    messageHandler.on('EditAnnotation', (data) => {
      debugger
      const { annotation } = data
      if (annotation.rect) {
        setAnnotRect(annotation)
      }
      if (annotation.color) {
        setAnnotRGBColor(annotation)
      }
      if (annotation.transparency) {
        setAnnotTransparency(annotation)
      }
      if (annotation.content) {
        setAnnotContent(annotation)
      }
      if (annotation.linePoints) {
        setLinePoint(annotation)
      }
      if (annotation.inkPointes) {
        setInkPointes(annotation)
      }
      if (annotation.destPage) {
        setLinkDest(annotation)
      }
      if (annotation.url) {
        setLinkUri(annotation)
      }
    })

    messageHandler.on('FlattenPage', (data) => {
      const { doc, pagesPtr, saveType } = data
      console.log(pagesPtr)
      for (let i = 0; i < pagesPtr.length; i++) {
        Module._FlattenPage(pagesPtr[i].pagePtr)
      }
      return saveDocumentByStream({
        doc,
        saveType
      })
    })

    messageHandler.on('CreateAnnotation', (data) => {
      const { doc, pagePtr, annotation } = data
      return createAnnotation(doc, pagePtr, annotation)
    })

    messageHandler.on('RemoveAnnot', (data) => {
      const { annotPtr } = data
      const res = Module._RemoveAnnot(annotPtr)
      console.log(res)
    })

    messageHandler.on('XFDFExportAnnotations', (data) => {
      const { doc } = data
      const xfdfStream = exportXFDFByStream(doc)
      return xfdfStream.Content
    })

    messageHandler.on('XFDFImportAnnotations', (data) => {
      const { doc, xfdfBuffer } = data

      let xfdfBufferPtr = _malloc(xfdfBuffer.length)
      Module.HEAPU8.set(xfdfBuffer, xfdfBufferPtr)
debugger
      const res = Module._XFDFImportAnnotaitons(doc, xfdfBufferPtr, xfdfBuffer.length)
      console.log(res)
    })

    messageHandler.on('SaveDocumentByStream', (data) => {
      console.log('SaveDocumentByStream')
      return saveDocumentByStream(data)
    })

    messageHandler.on('SetPassword', (data) => {
      const { doc, password: rawPassword } = data
      const password = stringToNewUTF8(rawPassword)
      const res =  Module._SetPassword(doc, password, '', '', '', '', '', '')
    })

    messageHandler.on('InitEditPage', (data) => {
      const { doc, pageIndex } = data
      const pagePtr = Module._GetPage(doc, pageIndex)
      const editPagePtr = Module._InitEditPage(pagePtr)
      return { editPagePtr, pagePtr }
    })

    messageHandler.on('BeginEdit', (data) => {
      const { editPagePtr, type } = data
      Module._BeginEdit(editPagePtr, type)
    })

    messageHandler.on('EndEdit', (editPagePtr) => {
      Module._EndEdit(editPagePtr)
    })

    messageHandler.on('GetEditAreaCount', (editPagePtr) => {
      return Module._GetEditAreaCount(editPagePtr)
    })

    messageHandler.on('GetEditArea', (data) => {
      const { editPagePtr, index } = data
      return Module._GetEditArea(editPagePtr, index)
    })

    messageHandler.on('IsTextArea', (editAreaPtr) => {
      return Module._IsTextArea(editAreaPtr)
    })

    messageHandler.on('SetFrame', (data) => {
      const { pagePtr, editAreaPtr, rect } = data

      Module._SetFrame(
        pagePtr,
        editAreaPtr,
        rect.left,
        rect.right,
        rect.bottom,
        rect.top
      )
    })

    messageHandler.on('GetFrame', (data) => {
      const { pagePtr, editAreaPtr } = data

      Rect = {}

      Module._GetFrame(pagePtr, editAreaPtr)

      return {
        left: Rect.Left,
        top: Rect.Top,
        right: Rect.Right,
        bottom: Rect.Bottom
      }
    })

    messageHandler.on('GetBeginAndEndCharPlace', (editAreaPtr) => {
      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetBeginAndEndCharPlace(editAreaPtr)

      return {
        start: EditCharPlace,
        end: EndEditCharPlace
      }
    })

    messageHandler.on('GetTextByRange', (data) => {
      const { editAreaPtr, start, end } = data
      let result = Module._GetTextByRange(
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex,
      )
      let text = UTF8ToString(result)
      return text
    })

    messageHandler.on('GetText', async (editAreaPtr) => {

      EditCharPlace = {}
      EndEditCharPlace = {}

      await Module._GetBeginAndEndCharPlace(editAreaPtr)

      let result = await Module._GetTextByRange(
        editAreaPtr,
        EditCharPlace.SectionIndex,EditCharPlace.LineIndex,EditCharPlace.RunIndex,EditCharPlace.CharIndex,
        EndEditCharPlace.SectionIndex,EndEditCharPlace.LineIndex,EndEditCharPlace.RunIndex,EndEditCharPlace.CharIndex,
      )
      let text = UTF8ToString(result)
      return text
    })

    messageHandler.on('SelectCharsRangeAtPos', (data) => {
      const { pagePtr, editAreaPtr, startPoint, endPoint } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._SelectCharsRangeAtPos(
        pagePtr,
        editAreaPtr,
        startPoint.x,startPoint.y,
        endPoint.x,endPoint.y,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetSectionAtPos', (data) => {
      const { pagePtr, editAreaPtr, pointX, pointY } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetSectionAtPos(
        pagePtr,
        editAreaPtr,
        pointX,
        pointY
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetTextCursorPoints', (data) => {
      const { pagePtr, editAreaPtr, char } = data

      CursorPoints = {}

      Module._GetTextCursorPoints(
        pagePtr,
        editAreaPtr,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex,
      )
      return CursorPoints
    })

    messageHandler.on('GetPrevCharPlace', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetPrevCharPlace(
        editAreaPtr,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetNextCharPlace', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetNextCharPlace(
        editAreaPtr,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetUpCharPlace', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetUpCharPlace(
        editAreaPtr,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetDownCharPlace', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetDownCharPlace(
        editAreaPtr,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('InsertText', (data) => {
      const { editAreaPtr, char, text } = data

      EditCharPlace = {}

      let textPtr = stringToNewUTF8(text)

      Module._InsertText(
        editAreaPtr,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex,
        textPtr,
        false
      )
      return EditCharPlace
    })

    messageHandler.on('DeleteChar', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}

      Module._DeleteChar(
        editAreaPtr,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex,
        true
      )
      return EditCharPlace
    })

    messageHandler.on('DeleteChars', (data) => {
      const { editAreaPtr, start, end } = data

      EditCharPlace = {}

      Module._DeleteChars(
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex,
        true
      )
      return EditCharPlace
    })

    messageHandler.on('GetSelectLineRectsCounts', (data) => {
      const { editAreaPtr, start, end } = data
      return Module._GetSelectLineRectsCounts(
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex,
      )
    })

    messageHandler.on('GetSelectLineRects', (data) => {
      const { pagePtr, editAreaPtr, start, end } = data

      RectArray = []

      Module._GetSelectLineRects(
        pagePtr,
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex
      )
      return RectArray
    })

    messageHandler.on('SetCharsFontColor', (data) => {
      const { editAreaPtr, start, end, color } = data
      Module._SetCharsFontColor(
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex,
        color.r,color.g,color.b
      )
    })

    messageHandler.on('SetCharsFontTransparency', (data) => {
      const { editAreaPtr, start, end, opacity } = data
      Module._SetCharsFontTransparency(
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex,
        opacity
      )
    })

    messageHandler.on('SetTextAligningSection', (data) => {
      const { editAreaPtr, char, alignType } = data
      Module._SetTextAligningSection(
        editAreaPtr,
        alignType,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex
      )
    })

    messageHandler.on('SetTextAligningRange', (data) => {
      const { editAreaPtr, start, end, alignType } = data
      Module._SetTextAligningRange(
        editAreaPtr,
        alignType,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex
      )
    })

    messageHandler.on('SetCharsFontSize', (data) => {
      const { editAreaPtr, start, end, fontSize } = data
      Module._SetCharsFontSize(
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex,
        fontSize,
        true
      )
    })

    messageHandler.on('SetFontFromNativeTrueTypeFont', (data) => {
      const { editAreaPtr, start, end, fontFamily } = data

      const font = stringToNewUTF8(fontFamily)
      Module._SetFontFromNativeTrueTypeFont(
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex,
        font
      )
    })

    messageHandler.on('SetCharsFontStyle', async (data) => {
      const { editAreaPtr, start, end, fontStyle } = data

      let action = '_' + fontStyle

      Module[action](
        editAreaPtr,
        start.SectionIndex,start.LineIndex,start.RunIndex,start.CharIndex,
        end.SectionIndex,end.LineIndex,end.RunIndex,end.CharIndex
      )
    })

    messageHandler.on('CreateNewTextArea', (data) => {
      const { pagePtr, editPagePtr, rect, fontData, alignType } = data

      return Module._CreateNewTextArea(
        pagePtr,
        editPagePtr,
        rect.left,
        rect.right,
        rect.bottom,
        rect.top,
        stringToNewUTF8(fontData.fontName),
        fontData.fontSize,
        fontData.r,
        fontData.g,
        fontData.b,
        fontData.opacity,
        fontData.isBold,
        fontData.italic,
        alignType
      )
    })

    messageHandler.on('RemoveEditAreaByIndex', (data) => {
      const { editPagePtr, editAreaIndex } = data
      Module._RemoveEditAreaByIndex(
        editPagePtr,
        editAreaIndex
      )
    })

    messageHandler.on('RemoveEditArea', (data) => {
      const { editPagePtr, editAreaPtr } = data
      Module._RemoveEditArea(
        editPagePtr,
        editAreaPtr
      )
    })

    messageHandler.on('CopyArea', (data) => {
      const { editPagePtr, editAreaPtr } = data
      return Module._CopyArea(
        editPagePtr,
        editAreaPtr
      )
    })

    messageHandler.on('GetCopyAreaContent', (data) => {
      const { editPagePtr, editAreaInfoPtr } = data
      let result = Module._GetCopyAreaContent(
        editPagePtr,
        editAreaInfoPtr
      )
      let text = UTF8ToString(result)
      return text
    })

    messageHandler.on('GetTextStyle', (data) => {
      const { editAreaPtr, char } = data

      EditTextStyle = {}

      Module._GetTextStyle(
        editAreaPtr,
        char.SectionIndex,char.LineIndex,char.RunIndex,char.CharIndex,
      )
      return EditTextStyle
    })
  }

}

async function initDoc() {
  while (!runtimeInitialized) {
    await sleep(100)
  }
  return new Promise((resolve, reject) => {
    const doc = Module._InitDocument()
    resolve(doc)
  })
}

initDoc().then((doc) => {
  CPDFWorker.setup(doc)
})

function RenderPageBitmapWithMatrix(data) {
  const { pagePtr, scale, left, right, bottom, top, mode, flag, form } = data
  let pixelNum = parseInt(right - left) * parseInt(bottom - top)
  let imageBytes = pixelNum * 4
  let imageptr = _malloc(imageBytes)
  for (var i = 0; i < pixelNum; i++) {
    Module.HEAP32[imageptr / 4 + i] = 0
  }

  Module._RenderPageBitmapWithMatrix(pagePtr, scale, left, right, bottom, top, mode, imageptr, flag, form)
  let imageArray = new Uint8Array(imageBytes)
  for (var i = 0; i < imageBytes; i += 4) {
    //bgra 转 rgba
    imageArray[i] = Module.HEAPU8[imageptr + i + 2]
    imageArray[i + 1] = Module.HEAPU8[imageptr + i + 1]
    imageArray[i + 2] = Module.HEAPU8[imageptr + i]
    imageArray[i + 3] = Module.HEAPU8[imageptr + i + 3]
  }
  _free(imageptr)

  return imageArray
}

function createAnnotation(doc, pagePtr, annotation) {
  const typeInt = AnnotationType[annotation.type.toUpperCase()]

  const annotPtr = Module._CreateAnnot(doc, pagePtr, typeInt)

  const { color, lineWidth, content, createDate, transparency, fillTransparency } = annotation
  if (color) {
    setAnnotRGBColor({
      annotPtr,
      color
    })
  }

  let outline=  _NewOutlineRoot(doc)
  let titleUrl="我赛待定ABCDERFHIJklmnopqrstuvwxyz1"
  //let titlePtr= stringToNewUTF8(titleUrl)
  let titlePtr=stringToNewUTF8(titleUrl)
  _SetTitle(outline,titlePtr)
  let result_title=_GetTitle(outline)
  let title_text=UTF8ToString(result_title)
  console.log(title_text)

  if (content) {
    setAnnotContent({
      annotPtr,
      content
    })
  }

  if (createDate) {
    setAnnotCreationDate({
      annotPtr,
      createDate
    })
  }

  if (transparency) {
    setAnnotTransparency({
      annotPtr,
      transparency
    })
  }

  if (fillTransparency) {
    SetAnnotFilledTransparency({
      annotPtr,
      fillTransparency
    })
  }
  

  switch (typeInt) {
    case AnnotationType.TEXT:
      break
    case AnnotationType.LINK:
      break
    case AnnotationType.FREETEXT:
      const { bgColor, fontName, fontSize, textAlignment } = annotation
      if (bgColor) {
        Module._SetFreeTextBgColor(annotPtr, bgColor.R / 255, bgColor.G / 255, bgColor.B / 255)
      } else {
        Module._ClearFreeTextBgColor(annotPtr)
      }

      setFreeText({
        annotPtr,
        fontName,
        fontSize,
        color,
        textAlignment,
        content
      })
      break
    case AnnotationType.LINE:
      const { tail, linePoints } = annotation
      if (tail) {
        Module._SetLineType(annotPtr, 0, 1)
      }
      Module._SetLinePoint(pagePtr, annotPtr, linePoints[0], linePoints[1], linePoints[2], linePoints[3])
    case AnnotationType.SQUARE:
    case AnnotationType.CIRCLE:
      createShape({
        annotPtr,
        color,
        lineWidth,
        transparency,
        fillTransparency
      })
      break
    case AnnotationType.HIGHLIGHT:
    case AnnotationType.UNDERLINE:
    case AnnotationType.SQUIGGLY:
    case AnnotationType.STRIKEOUT:
      const { quadPoints } = annotation
      createMarkup({
        pagePtr,
        annotPtr,
        color,
        transparency,
        quadPoints
      })
      break
    case AnnotationType.STAMP:
      const { stampType } = annotation
      if (stampType === 'standard') {
        createStandardStamp({
          annotPtr,
          content
        })
      } else if (stampType === 'text') {
        annotation.annotPtr = annotPtr
        createTextStamp(annotation)
      }
      break
    case AnnotationType.INK:
      const { inkPointes } = annotation
      setAnnotBorderWidth({
        annotPtr,
        lineWidth
      })
      setInkPointes({
        pagePtr,
        annotPtr,
        inkPointes
      })
  }
  const { left, right, top, bottom } = annotation.rect
  Module._SetAnnotRect(pagePtr, annotPtr, left, right, bottom, top)

  Module._UpdateAnnotAp(annotPtr, 0)
  return annotPtr
}

function getCommonAnntationAttr({
  pagePtr,
  annotPtr
}) {
  const lineWidth = Module._GetAnnotBorderWidth(annotPtr)
  const annotation = {
    lineWidth,
  }
  
  return annotation
}

function getMarkup({
  pagePtr,
  annotPtr
}) {
  RGBColor = {}
  const annotationAttr = {}
  const color = {}
  const hasColor = Module._GetAnnotRGBColor(annotPtr)
  if (hasColor) {
    for (key in RGBColor) {
      color[key] = roundToDecimalPlaces(RGBColor[key] * 255)
    }
    annotationAttr.color = color
  }

  InkPointArray = []
  Module._GetQuadPoints(pagePtr, annotPtr)
  annotationAttr.quadPoints = InkPointArray

  return annotationAttr
}

function getInk({
  pagePtr,
  annotPtr
}) {
  const count = Module._GetInkPathCount(annotPtr)

  const inkPointes = []
  for (let i = 0; i < count; i++) {
    InkPointArray = []
    Module._GetInkPoints(pagePtr, annotPtr, i)
    inkPointes.push(InkPointArray)
  }

  const lineWidth = Module._GetAnnotBorderWidth(annotPtr)

  return {
    lineWidth,
    inkPointes
  }
}

function getAnnotation({
  doc,
  pagePtr,
  annotPtr,
  typeInt,
  scale
}) {
  let annotationAttr = null
  Rect = {}
  Module._GetAnnotRect(pagePtr, annotPtr)
  const rect = {
    left: roundToDecimalPlaces(Rect.Left),
    top: roundToDecimalPlaces(Rect.Top),
    right: roundToDecimalPlaces(Rect.Right),
    bottom: roundToDecimalPlaces(Rect.Bottom)
  }

  const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
  const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))
  const rawContent = Module._GetAnnotContent(annotPtr)
  const content = UTF8ToString(rawContent)

  const transparency = roundToDecimalPlaces(Module._GetAnnotTransparency(annotPtr))

  const annotation = {
    createDate,
    content,
    rect,
    transparency
  }

  let hasColor = 0
  let isFilled = 0
  let hasBgColor = 0
  let fillTransparency = null
  switch (typeInt) {
    case AnnotationType.TEXT:
      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      annotation.color = {
        R: roundToDecimalPlaces(RGBColor.R * 255),
        G: roundToDecimalPlaces(RGBColor.G * 255),
        B: roundToDecimalPlaces(RGBColor.B * 255)
      }

      return annotation
    case AnnotationType.LINK:
      const actionType = Module._GetLinkActionType(annotPtr)
      annotation.actionType = ActionTypeString[actionType]
      RGBColor = {}
      isFilled = Module._GetAnnotFilledRGBColor(annotPtr)
      if (isFilled) {
        const fillColor = {}
        for (key in RGBColor) {
          fillColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }
        annotation.fillColor = fillColor
      }
      if (actionType === 1) {
        PDFDestination = {}
        Module._GetLinkDestination(doc, annotPtr)
        annotation.destPage = PDFDestination.PageIndex + 1
        annotation.url = ''
      } else if (actionType === 6) {
        const rawLinkUri = Module._GetLinkUri(annotPtr)
        const linkUri = UTF8ToString(rawLinkUri)
        annotation.url = linkUri
        annotation.destPage = ''
      }
      return annotation
    case AnnotationType.FREETEXT:
      FontDa = {}
      const rawFontName = Module._GetFreeTextFont(annotPtr)
      const fontName = UTF8ToString(rawFontName)
      annotation.fontName = fontName
      annotation.fontSize = FontDa.fontSize

      const textAlignment = Module._GetTextAlignment(annotPtr)
      annotation.textAlignment = textAlignment

      annotation.color = {
        R: roundToDecimalPlaces(FontDa.R * 255),
        G: roundToDecimalPlaces(FontDa.G * 255),
        B: roundToDecimalPlaces(FontDa.B * 255),
        A: FontDa.A
      }

      const rawFreetextContent = Module._GetFreeTextContent(annotPtr)
      const freetextContent = UTF8ToString(rawFreetextContent)
      annotation.content = freetextContent

      RGBColor = {}
      hasBgColor = Module._GetFreeTextBgColor(annotPtr)
      if (hasBgColor) {
        const bgColor = {}
        for (key in RGBColor) {
          bgColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }
        annotation.bgColor = bgColor
      }
      return annotation
    case AnnotationType.LINE:
      LinePoint = {}
      Module._GetLinePoint(pagePtr, annotPtr)
      const linePoints = [LinePoint.StartPointX, LinePoint.StartPointY, LinePoint.EndPointX, LinePoint.EndPointY]
      annotation.linePoints = linePoints

      LineType = {}
      Module._GetLineType(annotPtr)
      annotation.tail = LineTypeString[LineType.Tail],
      annotation.head = LineTypeString[LineType.Head]

      RGBColor = {}
      hasColor = Module._GetAnnotRGBColor(annotPtr)
      if (hasColor) {
        const color = {}
        for (key in RGBColor) {
          color[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }

        annotation.color = color
      }

      annotation.lineWidth = Module._GetAnnotBorderWidth(annotPtr)
      return annotation
    case AnnotationType.SQUARE:
    case AnnotationType.CIRCLE:
      const lineWidth = Module._GetAnnotBorderWidth(annotPtr)

      annotation.lineWidth = lineWidth

      RGBColor = {}
      hasColor = Module._GetAnnotRGBColor(annotPtr)
      if (hasColor) {
        const color = {}
        for (key in RGBColor) {
          color[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }

        annotation.color = color
      }

      RGBColor = {}
      isFilled = Module._GetAnnotFilledRGBColor(annotPtr)
      if (isFilled) {
        const fillColor = {}
        for (key in RGBColor) {
          fillColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }
        annotation.fillColor = fillColor
      }

      fillTransparency = Module._GetAnnotFilledTransparency(annotPtr)
      annotation.fillTransparency = fillTransparency

      return annotation
    case AnnotationType.HIGHLIGHT:
    case AnnotationType.UNDERLINE:
    case AnnotationType.SQUIGGLY:
    case AnnotationType.STRIKEOUT:
      annotationAttr = getMarkup({
        pagePtr,
        annotPtr
      })
      Object.assign(annotation, annotationAttr)
      return annotation
    case AnnotationType.STAMP:
      const stampType = Module._GetStampType(annotPtr)
      console.log(stampType)
      annotation.stampType = StampTypeString[stampType]
      if (stampType === 0) {
        const width = parseInt((annotation.rect.right - annotation.rect.left) * scale + 1)
        const height = parseInt((annotation.rect.bottom - annotation.rect.top) * scale + 1)
        let pixelNum = width * height
        console.log(width, height, scale)
         let imageBytes = pixelNum * 4
         let imageptr = _malloc(imageBytes)
         for (var i = 0; i < pixelNum; i++) {
             Module.HEAP32[imageptr / 4 + i] = 0
         }

        Module._RenderAnnot(annotPtr, width, height, imageptr, 0)
        let imageArray = new Uint8Array(imageBytes)
        for (var i = 0; i < imageBytes; i += 4) {
          //bgra 转 rgba
          imageArray[i] = Module.HEAPU8[imageptr + i + 2]
          imageArray[i + 1] = Module.HEAPU8[imageptr + i + 1]
          imageArray[i + 2] = Module.HEAPU8[imageptr + i]
          imageArray[i + 3] = Module.HEAPU8[imageptr + i + 3]
        }

        annotation.imageArray = imageArray
        annotation.stampType = 'image'
      } else {
        TextStamp = {}
        Module._GetTextStamp(annotPtr)
        console.log(TextStamp)
        annotation.content = TextStamp.Content
        annotation.stampColor = TextStamp.StampColor
        annotation.stampShape = TextStampShapeString[TextStamp.StampType]
        if (TextStamp.Date) {
          annotation.date = TextStamp.Date
        }
      }
      return annotation
    case AnnotationType.INK:
      RGBColor = {}
      hasColor = Module._GetAnnotRGBColor(annotPtr)
      if (hasColor) {
        const color = {}
        for (key in RGBColor) {
          color[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }

        annotation.color = color
      }

      annotationAttr = getInk({
        pagePtr,
        annotPtr
      })
      Object.assign(annotation, annotationAttr)
      return annotation
  }

  return 
}

function getWidgetAnnotation({
  doc,
  pagePtr,
  annotPtr,
  typeInt
}) {
  // WidgetType
  let annotationAttr = null
  Rect = {}
  Module._GetAnnotRect(pagePtr, annotPtr)
  const rect = {
    left: roundToDecimalPlaces(Rect.Left),
    top: roundToDecimalPlaces(Rect.Top),
    right: roundToDecimalPlaces(Rect.Right),
    bottom: roundToDecimalPlaces(Rect.Bottom)
  }

  const rawCreateDate = Module._GetAnnotCreationDate(annotPtr)
  const createDate = parseAdobePDFTimestamp(UTF8ToString(rawCreateDate))
  const rawContent = Module._GetAnnotContent(annotPtr)
  const content = UTF8ToString(rawContent)


  Module._GetWidgetBorderRGBColor(annotPtr)

  const type = Module._GetWidgetType(annotPtr)

  const annotation = {
    createDate,
    content,
    rect,
    type: WidgetTypeString[type]
  }

  if (type === 1) {
    const checkStyle = Module._GetWidgetCheckStyle(annotPtr)
    annotation.checkStyle = CheckStyleString[checkStyle]

    RGBColor = {}
    const hasColor = Module._GetWidgetCheckRGBColor(annotPtr)
    if (hasColor) {
      const checkRGBColor = {
        R: roundToDecimalPlaces(RGBColor.R * 255),
        G: roundToDecimalPlaces(RGBColor.G * 255),
        B: roundToDecimalPlaces(RGBColor.B * 255)
      }
      annotation.checkRGBColor = checkRGBColor
    }
  }

  if (type === 3) {
    const textFieldSpe = Module._GetTextFieldSpe(annotPtr)
    if (textFieldSpe === 1) {
      return null
    }

    const rawContent = Module._GetTextFieldContent(annotPtr)
    const content = UTF8ToString(rawContent)
    annotation.content = content

    const isMultiLine = Module._IsMultiLine(annotPtr)
    annotation.isMultiLine = isMultiLine

    
    const align = Module._GetWidgetTextJustification(annotPtr)
    annotation.textAlignment = ALIGN[align]
  }

  if (type === 0) {
    const rawTitle = Module._GetButtonTitle(annotPtr)
    const title = UTF8ToString(rawTitle)
    annotation.title = title

    const actionType = Module._GetBtnActionType(annotPtr)

    const rawActionName = Module._GetBtnActionNamedAction(annotPtr)
    const actionName = UTF8ToString(rawActionName)
    console.log(actionName)
    annotation.actionName = actionName
    if (actionType === 1) {
      const destPage = Module._GetBtnActionDestPage(doc, annotPtr)
      annotation.destPage = destPage + 1

    } else if (actionType === 6) {
      const rawActionURI = Module._GetBtnActionURI(annotPtr)
      const actionURI = UTF8ToString(rawActionURI)
      annotation.url = actionURI
    }

  }
  if (type === 4 || type === 5) {
    const rawSelectedValue = Module._GetItemsSelectedValue(annotPtr)
    const selectedValue = UTF8ToString(rawSelectedValue)
    annotation.selectedValue = selectedValue

    WidgetItem = {}
    Module._GetItemsSelectedItem(annotPtr)
    annotation.selectedItem = {
      values: WidgetItem.Values,
      strings: WidgetItem.Strings
    }

    const itemsCount = Module._GetItemsCount(annotPtr)
    annotation.itemsCount = itemsCount
    WidgetItemsArray = []
    Module._GetWidgetItems(annotPtr)
    annotation.items = WidgetItemsArray
  }
  const checkBoxIsChecked = Module._GetCheckBoxIsChecked(annotPtr)
  annotation.checkBoxIsChecked = checkBoxIsChecked
  const radioButtonIsChecked = Module._GetRadioButtonIsChecked(annotPtr)
  annotation.radioButtonIsChecked = radioButtonIsChecked

  RGBColor = {}
  const hasColor = Module._GetWidgetBorderRGBColor(annotPtr)
  if (hasColor) {
    const bgColor = {
      R: roundToDecimalPlaces(RGBColor.R * 255),
      G: roundToDecimalPlaces(RGBColor.G * 255),
      B: roundToDecimalPlaces(RGBColor.B * 255)
    }
    annotation.bgColor = bgColor
  }

  RGBColor = {}
  const hasBorderColor = Module._GetWidgetBorderRGBColor(annotPtr)
  if (hasBorderColor) {
    const borderColor = {
      R: roundToDecimalPlaces(RGBColor.R * 255),
      G: roundToDecimalPlaces(RGBColor.G * 255),
      B: roundToDecimalPlaces(RGBColor.B * 255)
    }
    annotation.borderColor = borderColor
  }

  const borderStyleType = Module._GetWidgetBorderStyle(annotPtr)
  annotation.borderStyle = BorderStyleString[borderStyleType]
  return annotation
}

function saveDocumentByStream(data) {
  const { doc, saveType = 2 } = data
  console.log(saveType)
  DataArray=[]
  Module._SaveDocumentByStream(doc, saveType);
  let pdfData = []
  for(let i = 0; i < DataArray.length; i++) {
    let subArray = new Uint8Array(DataArray[i])
    pdfData.push(subArray)
  }
  DataArray=[]
  return pdfData
}

function exportXFDFByStream(doc) {
  XFDFData={}
  console.log(doc)
  Module._XFDFExportAnnotations(doc)
  console.log(XFDFData)
  return XFDFData
}

function setAnnotCreationDate(data) {
  const { annotPtr, createDate: rawCreateDate } = data

  const createDate = stringToNewUTF8(convertToPDFTimestamp(rawCreateDate))
  Module._SetAnnotCreationDate(annotPtr, createDate)
}

function setAnnotRect(data) {
  const { pagePtr, annotPtr, rect } = data
  const { left, right, bottom, top } = rect

  const result = Module._SetAnnotRect(pagePtr, annotPtr, left, right, bottom, top)
  return result
}

function setAnnotContent(data) {
  const { annotPtr, content: rawContent } = data

  debugger
  const content = stringToNewUTF8(rawContent)
  const result = Module._SetAnnotContent(annotPtr, content)

  const contentString = Module._GetAnnotContent(annotPtr)
  const res = UTF8ToString(contentString)
  console.log(res)

  return result
}

function setAnnotRGBColor(data) {
  const { annotPtr, color } = data
  const { R = 255, G = 255, B = 255 } = color

  const isSetColorSuccess = Module._SetAnnotRGBColor(annotPtr, R / 255, G / 255, B / 255)
  return !!isSetColorSuccess
}

function setAnnotTransparency(data) {
  const { annotPtr, transparency = 1 } = data

  const result = Module._SetAnnotTransparency(annotPtr, transparency)
  return result
}

function setAnnotBorderWidth(data) {
  const { annotPtr, lineWidth } = data

  Module._SetAnnotBorderWidth(annotPtr, lineWidth)
}

function SetAnnotFilledTransparency(data) {
  const { annotPtr, transparency = 1 } = data
  const result = Module._SetAnnotFilledTransparency(annotPtr, transparency)
  return result
}

function setQuadPoints(data) {
  const { pagePtr, annotPtr, quadPoints } = data
  const quadPointsArray = []
  for (let i = 0; i < quadPoints.length; i++) {
    const quadPoint = quadPoints[i]
    quadPointsArray.push(quadPoint.PointX, quadPoint.PointY)
  }

  const quadPointsArrayLength = quadPointsArray.length
  let quadPointsArrayBytes = quadPointsArrayLength * 4
  let quadPointsArrayPtr = _malloc(quadPointsArrayBytes)
  for (var i = 0; i < quadPointsArrayLength; i++) {
    Module.HEAPF32[quadPointsArrayPtr / 4 + i] = quadPointsArray[i]
  }

  const result = Module._SetQuadPoints(pagePtr, annotPtr, quadPointsArrayPtr, quadPointsArrayLength)
  _free(quadPointsArrayPtr)
  return result
}

function setLinePoint(data) {
  const { pagePtr, annotPtr, linePoints } = data

  const result = Module._SetLinePoint(pagePtr, annotPtr, linePoints[0], linePoints[1], linePoints[2], linePoints[3])
  return result
}

function setInkPointes(data) {
  const { pagePtr, annotPtr, inkPointes } = data

  const inkPointesArray = []
  const inklistArray = []
  const lineLength = inkPointes.length
  for (let i = 0; i < lineLength; i++) {
    const length = inkPointes[i].length
    for (let j = 0; j < length; j++) {
      inkPointesArray.push(inkPointes[i][j].PointX, inkPointes[i][j].PointY)
    }
    inklistArray.push(length)
  }

  const inklistArrayLength = inklistArray.length
  let inklistArrayBytes = inklistArrayLength * 4
  let inklistArrayPtr = _malloc(inklistArrayBytes)
  for (var i = 0; i < inklistArrayLength; i++) {
    Module.HEAP32[inklistArrayPtr / 4 + i] = inklistArray[i]
  }

  const inkPointesArrayLength = inkPointesArray.length
  let inkPointesArrayBytes = inkPointesArrayLength * 4
  let inkPointesArrayPtr = _malloc(inkPointesArrayBytes)
  for (var i = 0; i < inkPointesArrayLength; i++) {
    Module.HEAPF32[inkPointesArrayPtr / 4 + i] = inkPointesArray[i]
  }

  const res = Module._SetInkPath(pagePtr, annotPtr, inklistArrayPtr, lineLength, inkPointesArrayPtr)
  console.log(res)
}

function setLinkDest(data) {
  const { doc, annotPtr, destPage, pageHeight } = data
  Module._SetLinkDestination(doc, annotPtr, destPage - 1, 0, pageHeight, 0)
}

function setLinkUri(data) {
  const { annotPtr, url: rawUrl } = data
  const url = stringToNewUTF8(rawUrl)
  const res = Module._SetLinkUri(annotPtr, url)
  console.log(res)
}

function createMarkup(data) {
  const { pagePtr, annotPtr, color, transparency, quadPoints  } = data
  setAnnotRGBColor({
    annotPtr,
    color
  })

  setAnnotTransparency({
    annotPtr,
    transparency
  })

  setQuadPoints({
    pagePtr,
    annotPtr,
    quadPoints
  })
}

function setFreeText(data) {
  const { annotPtr, fontName: rawFontName, fontSize, color, textAlignment, content: rawContent } = data

  const fontName = stringToNewUTF8(rawFontName)
  const content = stringToNewUTF8(rawContent)

  Module._SetFreeText(annotPtr, fontName, fontSize, color.R / 255, color.G / 255, color.B / 255, textAlignment, content)
}

function createShape(data) {
  const { annotPtr, lineWidth } = data
  setAnnotBorderWidth({
    annotPtr,
    lineWidth
  })
}

function createStandardStamp(data) {
  const { annotPtr, content: rawContent } = data

  const content = stringToNewUTF8(rawContent)
  Module._SetStandardStamp(annotPtr, content, 0)
}

function createTextStamp(data) {
  const { annotPtr, content: rawContent, date: rawDate, stampShape, stampColor } = data

  const content = stringToNewUTF8(rawContent)
  const date = stringToNewUTF8(rawDate)

  Module._SetTextStamp(annotPtr, content, date, stampShape, stampColor, 0)
}
